#!/usr/bin/env python
# -*- coding: UTF-8 -*-

"""
    Copyright (C) 2009 Mitsuya Shibata

    This file is part of qr-creator.

    qr-creator is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    qr-creator is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with qr-creator.  If not, see <http://www.gnu.org/licenses/>.
"""


APP_NAME = "qr-creator"
LOCALE_DIR = "/usr/share/locale"

import sys, os
try:
	import pygtk
	pygtk.require('2.0')
except:
	pass
try:
	import gtk
	import gtk.glade
	import pango
	import cairo
except:
	sys.exit(1)

# i18nのために必要
import locale
import gettext

# PNG画像へ変換するために必要
import Image, ImageDraw

# バージョン決定に使う：data[バージョン][ECC] = データコード語数
data = [
	[19, 16, 13, 9], [34, 28, 22, 16], [55, 44, 34, 26], [80, 64, 48, 36], [108, 86, 62, 46],
	[136, 108, 76, 60], [156, 124, 88, 66], [194, 154, 110, 86], [232, 182, 132, 100], [274, 216, 154, 122],
	[324, 254, 180, 140], [370, 290, 206, 158], [428, 334, 244, 180], [461, 365, 261, 197], [523, 415, 295, 223],
	[589, 453, 325, 253], [647, 507, 367, 283], [721, 563, 397, 313], [795, 627, 445, 341], [861, 669, 485, 385],
	[932, 714, 512, 406], [1006, 782, 568, 442], [1094, 860, 614, 464], [1174, 914, 664, 514], [1276, 1000, 718, 538],
	[1370, 1062, 754, 596], [1468, 1128, 808, 628], [1531, 1193, 871, 661], [1631, 1267, 911, 701], [1735, 1373, 985, 745],
	[1843, 1455, 1033, 793], [1955, 1541, 1115, 845], [2071, 1631, 1171, 901], [2191, 1725, 1231, 961], [2306, 1812, 1286, 986],
	[2434, 1914, 1354, 1054], [2566, 1992, 1426, 1096], [2702, 2102, 1502, 1142], [2812, 2216, 1582, 1222], [2956, 2334, 1666, 1276],
	]

class GtkFrontend():
	"""コマンドラインオプションを受け取り、メインウィンドウを作成する。"""

	def __init__(self, opt):
		"""初期設定を行い、ウィンドウを作成する。

		opt: コマンドラインオプション"""

		# ロケールとgettextの設定
		locale.setlocale(locale.LC_ALL, "")
		for module in gtk.glade, gettext:
			module.bindtextdomain(APP_NAME, LOCALE_DIR)
			module.textdomain(APP_NAME)
		import __builtin__
		__builtin__._ = gettext.gettext

		# gladeファイルを読み込みと同時にオブジェクトの作成
		# NOTE: Glade-3で作ったファイルをUIファイルに変換（暫定処置）
		#  gtk-builder-convert qrcreator.glade qrcreator.ui
		self.builder = gtk.Builder()
		self.gladefile = os.path.join("/usr/share/qr-creator", "qrcreator.ui")
		self.builder.add_from_file(self.gladefile)

		# イベントとハンドラの関連付け
		dic = {"on_MainWindow_destroy" : gtk.main_quit,
			"on_CreateButton_clicked" : self.createbutton_clicked,
			"on_SaveButton_clicked" : self.savebutton_clicked,
			"on_VerSpinButton_value_changed" : self.change_option,
			"on_DotSpinButton_value_changed" : self.change_option,
			"on_EccComboBox_changed" : self.change_option,
			"on_TextView_key_release_event" : self.change_textview,
			"on_DrawingArea_expose_event" : self.drawingarea_expose_event}
		self.builder.connect_signals(dic)

		# メインウィンドウの取得
		self.window = self.builder.get_object("MainWindow")
		self.window.show()

		# 文字数ラベルの前景色取得
		self.fg = self.builder.get_object("CharNumLabel").get_style().fg[gtk.STATE_NORMAL]

		# ECCの初期値設定
		ecc = ["L", "M", "H", "Q"]
		if opt.ecc in ecc:
			self.builder.get_object("EccComboBox").set_active(ecc.index(opt.ecc))
		else:
			self.builder.get_object("EccComboBox").set_active(1)

		# ドットサイズの初期値設定
		spinrange = self.builder.get_object("DotSpinButton").get_range()
		if opt.dot >= spinrange[0] and opt.dot <= spinrange[1]:
			self.builder.get_object("DotSpinButton").set_value(opt.dot)

		# バージョンの初期値設定
		spinrange = self.builder.get_object("VerSpinButton").get_range()
		if opt.width >= spinrange[0] and opt.width <= spinrange[1]:
			self.builder.get_object("VerSpinButton").set_value(opt.width)

		# 文字列の初期値設定
		if opt.string != None:
			textbuffer = gtk.TextBuffer()
			textbuffer.set_text(opt.string)
			self.builder.get_object("TextView").set_buffer(textbuffer)

		# 文字数の計算
		self.update_charnum()


		# DrawingAreaウィジェットの取得
		self.area = self.builder.get_object("DrawingArea")
		self.arearect = self.area.get_allocation()
		self.pixmap = gtk.gdk.Pixmap(self.area.window, self.arearect[2], self.arearect[3])
		self.pixmap.draw_rectangle(self.area.get_style().bg_gc[gtk.STATE_NORMAL],
		 True, 0, 0, self.arearect[2], self.arearect[3])
		self.cr = self.pixmap.cairo_create()

		# メインループ
		gtk.main()

	def createbutton_clicked(self, widget):
		"""入力データを元にコードを作成し、DrawingAreaに表示する。"""
		opt = QrOptions()
		opt = self.get_options()

		current, upper = self.calc_charnum(opt)
		if upper != 0 and current > upper:
			# 文字数が上限を越えているようなら、警告
			dlg = gtk.MessageDialog(None,
						gtk.DIALOG_MODAL,
						gtk.MESSAGE_WARNING,
						gtk.BUTTONS_CLOSE,
						_("Number of characters are more than upper limit. Please change ECC or version."))
			dlg.set_title(_("Warning"))
			res = dlg.run()
			dlg.destroy()
			return False

		# QRバーコードに変換
		qr = QrCreator(opt)
		bar = qr.str2qrdata(opt.string)

		# 幅とオフセットの取得
		width = len(bar[0]) * opt.dot
		offset = (self.arearect[2] - width) // 2

		# DrawingAreaのリセット
		self.pixmap.draw_rectangle(self.area.get_style().bg_gc[gtk.STATE_NORMAL],
		 True, 0, 0, self.arearect[2], self.arearect[3])

		# DrawingAreaに表示
		for i, line in enumerate(bar):
			for j, dot in enumerate(line):
				if dot == "1":
					self.cr.set_source_rgb(0, 0, 0)
				else:
					self.cr.set_source_rgb(1, 1, 1)
				x = j * opt.dot + offset
				y = i * opt.dot + offset
				self.cr.rectangle(x, y, opt.dot, opt.dot)
				self.cr.fill()

		# DrawingAreaの再描画
		self.area.queue_draw_area(0, 0, self.arearect[2], self.arearect[3])

	def savebutton_clicked(self, widget):
		"""入力データを元にコードを作成し、画像として保存する。"""
		opt = QrOptions()
		opt = self.get_options()

		current, upper = self.calc_charnum(opt)
		if upper != 0 and current > upper:
			# 文字数が上限を越えているようなら、警告
			dlg = gtk.MessageDialog(None,
						gtk.DIALOG_MODAL,
						gtk.MESSAGE_WARNING,
						gtk.BUTTONS_CLOSE,
						_("Number of characters are more than upper limit. Please change ECC or version."))
			dlg.set_title(_("Warning"))
			res = dlg.run()
			dlg.destroy()
			return False

		# QRバーコードに変換
		qr = QrCreator(opt)
		bar = qr.str2qrdata(opt.string)

		# ファイル名の初期化
		filename = "barcode.png"

		# ファイル選択ダイアログの種類
		dialog_action = gtk.FILE_CHOOSER_ACTION_SAVE
		# ファイル選択ダイアログのボタンの種類
		dialog_buttons = (gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL,
					gtk.STOCK_SAVE, gtk.RESPONSE_OK)

		# ファイル選択ダイアログの作成
		filedialog = gtk.FileChooserDialog(title=_("Save a file"),
						action=dialog_action,
						buttons=dialog_buttons)

		# 標準ファイル名の設定
		filedialog.set_current_name(filename)
		# 上書き時は確認する
		filedialog.set_do_overwrite_confirmation(True)
		# フィルタの設定
		filefilter = gtk.FileFilter()
		filefilter.set_name(_("Image files"))
		filefilter.add_pattern("*.png")
		filedialog.add_filter(filefilter)
		filefilter = gtk.FileFilter()
		filefilter.set_name(_("All files"))
		filefilter.add_pattern("*")
		filedialog.add_filter(filefilter)

		# ファイル選択ダイアログの表示
		if filedialog.run() == gtk.RESPONSE_OK:
			filename = filedialog.get_filename()
			filedialog.destroy()
		else:
			filedialog.destroy()
			return False

		qr.data2img(bar, filename)

	def get_options(self):
		"""現在の入力データを取得する。"""
		opt = QrOptions()

		textbuffer = self.builder.get_object("TextView").get_buffer()
		opt.string = textbuffer.get_text(textbuffer.get_start_iter(),
			textbuffer.get_end_iter())

		opt.dot = self.builder.get_object("DotSpinButton").get_value_as_int()
		opt.ecc = self.builder.get_object("EccComboBox").get_active_text()
		opt.version = self.builder.get_object("VerSpinButton").get_value_as_int()

		return opt

	def change_option(self, widget):
		"""オプションが更新された場合に、文字数の上限を再計算する。"""
		self.update_charnum()

	def change_textview(self, widget, event):
		"""TextViewが更新された場合に、文字数を再計算する。"""
		self.update_charnum()

	def update_charnum(self):
		"""現在の入力文字数と上限をラベルに反映させる。"""
		opt = QrOptions()
		opt = self.get_options()

		current, upper = self.calc_charnum(opt)

		label = str(current) + " / " + str(upper)
		if current > upper:
			# 上限を越えている場合は赤字で表示する
			self.builder.get_object("CharNumLabel").modify_fg(gtk.STATE_NORMAL,
				gtk.gdk.Color(65535, 0, 0))
		else:
			self.builder.get_object("CharNumLabel").modify_fg(gtk.STATE_NORMAL,
				self.fg)

		self.builder.get_object("CharNumLabel").set_text(label)

	def calc_charnum(self, opt):
		"""現在の入力文字数と、現在のオプションにおける上限を計算する。

		opt: オプション（QrOptions）
		戻り値: （現在の文字数, 上限）"""
		ecc = ["L", "M", "H", "Q"]

		# 文字列はShift JISで格納する必要がある
		current = len(opt.string.encode("cp932"))
		if opt.version == 0:
			# バージョン0の場合は、コード作成時に格納可能な
			# バージョンが自動的に指定されるので、規格上の最大値である
			# バージョン40の上限を指定する
			if opt.string.isdigit():
				# 数字のみで構成されている場合
				bit = (data[40 - 1][ecc.index(opt.ecc)] - 2) * 8
				upper = (bit // 10) * 3
				upper += (bit % 10) // 3
			else:
				# 8bitモードの場合
				upper = data[40 - 1][ecc.index(opt.ecc)] - 2
		else:
			if opt.string.isdigit():
				# 数字のみで構成されている場合
				bit = (data[opt.version - 1][ecc.index(opt.ecc)] - 2) * 8
				upper = (bit // 10) * 3
				upper += (bit % 10) // 3
			else:
				# 8bitモードの場合
				upper = data[opt.version - 1][ecc.index(opt.ecc)] - 2

		return (current, upper)


	def drawingarea_expose_event(self, widget, event):
		"""コード表示領域を再描画する。"""

		# 再描画イベントでは、ピックスマップを使う
		x, y, width, height = event.area
		widget.window.draw_drawable(widget.get_style().fg_gc[gtk.STATE_NORMAL],
		 self.pixmap, x, y, x, y, width, height)

		return False

class QrOptions():
	"""QRバーコードで扱うオプションを一括して管理する。"""
	def __init__(self):
		self.string = ""
		self.ecc = "M"
		self.dot = 1
		self.version = 0

class QrCreator():
	"""オプションに従ったQRバーコードを返す。"""

	def __init__(self, opt):
		"""
		オプションを解析し、QRバーコードを作成する。

		opt: オプション（辞書型）
		STR: 格納すべき文字列
		ECC: 誤り訂正レベル（L, M, H, Q）
		VER: バージョン（0-40、0の場合は自前で計算が必要）
		DOT: 1ドット（モジュール）あたりのピクセルサイズ

		設定不可
		MOD: モード（数字:1、英数字:2、8bit:3、漢字:4）
		"""

		# オプションの処理
		if opt.string != None:
			self.str = opt.string
		else:
			self.str = ""
		if opt.ecc != None:
			self.ecc = opt.ecc
		else:
			self.ecc = "M"
		if opt.version != None:
			self.ver = opt.version
		else:
			self.ver = 0
		if opt.dot != None:
			self.dot = opt.dot
		else:
			self.dot = 1

	def str2qrdata(self, string):
		"""与えられた文字列をQRバーコード列に変換する。

		戻り値: 変換されたバーコード"""

		self.str = string

		# 文字列の文字コードをShift JISに変換
		self.str = self.str.encode("cp932")

		# バージョン0なら適切なバージョンを計算
		if self.ver == 0:
			self.ver = self.calc_version()

		# コマンド文字列に変換
		path = os.path.join(os.path.dirname(__file__), "str2qr.pl")
		cmd = "perl " + path 
		cmd += " \"" + self.str
		cmd += "\" " + self.ecc
		cmd += " " + str(self.ver)
		cmd += " " + str(self.dot)

		# コマンドの実行
		p_stdin, p_stdout, p_stderr = os.popen3(cmd, "r")
		temp = p_stdout.read()
		err = p_stderr.read()
		if len(err) != 0:
			# エラー終了
			err = _("Coudn't convert to qr barcode.\n") + err
			sys.exit(err)

		# 出力をリストに変換
		barcode = temp.split("\n")

		return barcode

	def data2img(self, barcode, path):
		"""
		QRバーコード列をpathで指定されたファイルに保存する。

		barcode: 画像に変換したいQRバーコード列
		path: 保存するファイルの絶対パス付きファイル名
		"""

		# 画像のサイズ決定
		width = self.dot * len(barcode[0])
		height = width

		# グレイスケールモードで作成
		img = Image.new('L', (width, height), 255)
		canvas = ImageDraw.Draw(img)
		for i, line in enumerate(barcode):
			for j, dot in enumerate(line):
				if dot == "1":
					x = j * self.dot
					y = i * self.dot
					canvas.rectangle((x, y, x + self.dot - 1, y + self.dot -1 ), fill = 0)
		img.save(path, "PNG")

		return True

	def calc_version(self):
		"""データサイズから適切なバージョンを決定する。"""

		ver = 1

		# 必要ビット数の計算
		if self.str.isdigit():
			# 数字のみで構成されている場合
			bit = (len(self.str) // 3) * 10
			bit += (len(self.str) % 3) * 4
		else:
			# 8bitモードの場合
			bit = len(self.str) * 8

		ecc = 0
		if self.ecc == "L":
			ecc = 0
		elif self.ecc == "M":
			ecc = 1
		elif self.ecc == "Q":
			ecc = 2
		elif self.ecc == "H":
			ecc = 3

		# 格納可能なバージョンを確認
		for ver, line in enumerate(data):
			if bit <= (line[ecc] - 2) * 8:
				break
		ver += 1

		return ver
